<?php

use Drupal\block\Entity\Block;
use Drupal\block_content\BlockContentUuidLookup;
use Drupal\Core\Block\BlockPluginInterface;
use Drupal\Core\Field\FieldDefinitionInterface;
use Drupal\Core\Field\WidgetInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Render\Element;
use Drupal\Core\Url;

/**
 * Implements hook_form_FORM_ID_alter(): form_system_modules.
 */
function improvements_form_system_modules_alter(array &$form, FormStateInterface $form_state) {
  // Add autofocus to modules filter input
  if (isset($form['filters']['text'])) {
    $form['filters']['text']['#attributes']['autofocus'] = TRUE;
  }
}

/**
 * Implements hook_field_widget_third_party_settings_form().
 */
function improvements_field_widget_third_party_settings_form(WidgetInterface $widget, FieldDefinitionInterface $field_definition, string $form_mode, array $form, FormStateInterface $form_state) {
  $element = [];
  $widget_plugin_id = $widget->getPluginId();

  // "Long text" widgets
  if (in_array($widget_plugin_id, ['text_textarea', 'text_textarea_with_summary'])) {
    $element['max_length'] = [
      '#type' => 'number',
      '#title' => t('Max length'),
      '#default_value' => $widget->getThirdPartySetting('improvements', 'max_length'),
    ];

    $element['dynamic_height'] = [
      '#type' => 'checkbox',
      '#title' => t('Dynamic height'),
      '#default_value' => $widget->getThirdPartySetting('improvements', 'dynamic_height'),
    ];

    $element['disable_editor'] = [
      '#type' => 'checkbox',
      '#title' => t('Disable CKEditor'),
      '#default_value' => $widget->getThirdPartySetting('improvements', 'disable_editor'),
    ];

    $element['hide_format_select'] = [
      '#type' => 'checkbox',
      '#title' => t('Hide format select'),
      '#default_value' => $widget->getThirdPartySetting('improvements', 'hide_format_select'),
    ];
  }
  // "String" widget
  elseif ($widget_plugin_id == 'string_textfield') {
    $element['max_length'] = [
      '#type' => 'number',
      '#title' => t('Max length'),
      '#default_value' => $widget->getThirdPartySetting('improvements', 'max_length'),
    ];
  }
  // "File" widget
  elseif ($widget_plugin_id == 'file_generic' || $widget_plugin_id == 'image_image') {
    $element['autoupload'] = [
      '#type' => 'checkbox',
      '#title' => t('Auto upload'),
      '#default_value' => $widget->getThirdPartySetting('improvements', 'autoupload', TRUE),
    ];

    $element['upload_button_state'] = [
      '#type' => 'radios',
      '#title' => t('Upload button state'),
      '#options' => [
        2 => t('Visually Hidden'),
        1 => t('Visible'),
        0 => t('Without button'),
      ],
      '#default_value' => $widget->getThirdPartySetting('improvements', 'upload_button_state', 2),
    ];
  }

  return $element;
}

/**
 * Implements hook_field_widget_form_alter().
 */
function improvements_field_widget_form_alter(array &$element, FormStateInterface $form_state, array $context) {
  $widget = $context['widget']; /** @var WidgetInterface $widget */
  $widget_plugin_id = $widget->getPluginId();

  // Long text widgets
  if ($widget_plugin_id == 'text_textarea' || $widget_plugin_id == 'text_textarea_with_summary') {
    $third_party_settings = $widget->getThirdPartySettings('improvements');

    if (!empty($third_party_settings['max_length'])) {
      $element['#maxlength'] = $third_party_settings['max_length'];
    }
    if (!empty($third_party_settings['dynamic_height'])) {
      $element['#rows'] = improvements_get_textarea_rows_count((string)$element['#default_value'], (int)$element['#rows']);
    }
    if (!empty($third_party_settings['disable_editor'])) {
      $element['#editor'] = FALSE;
    }
    if (!empty($third_party_settings['hide_format_select'])) {
      $element['#after_build'][] = 'improvements_hide_format_select';
    }
  }
  // String widget
  elseif ($widget_plugin_id == 'string_textfield') {
    $third_party_settings = $widget->getThirdPartySettings('improvements');

    if (isset($third_party_settings['max_length'])) {
      $element['value']['#maxlength'] = $third_party_settings['max_length'];
    }
  }
  // File widget
  elseif ($widget_plugin_id == 'file_generic' || $widget_plugin_id == 'image_image') {
    $third_party_settings = $widget->getThirdPartySettings('improvements');

    $element['#autoupload'] = $third_party_settings['autoupload'] ?? TRUE;
    $element['#upload_button_state'] = $third_party_settings['upload_button_state'] ?? 2;
  }
  // Metatags widget
  elseif ($widget_plugin_id == 'metatag_firehose') {
    foreach (Element::children($element) as $key) {
      if (isset($element[$key]['#open'])) {
        $element[$key]['#open'] = FALSE;
      }
    }

    if (isset($element['basic'])) {
      $element['basic']['#open'] = TRUE;
    }
  }
}

/**
 * Return textarea rows count by text length.
 */
function improvements_get_textarea_rows_count(string $text, int $min_rows) {
  $rows = substr_count($text, "\n");
  $rows = min($rows + 1, 28);
  return max($rows, $min_rows);
}

/**
 * Formatted text element after build callback.
 */
function improvements_hide_format_select(array $element, FormStateInterface $form_state) {
  if (isset($element['format'])) {
    $element['format']['#access'] = FALSE;
  }

  return $element;
}

/**
 * Implements hook_form_FORM_ID_alter(): block_admin_display_form.
 */
function improvements_form_block_admin_display_form_alter(array &$form, FormStateInterface $form_state) {
  $block_content_uuid_lookup = \Drupal::service('block_content.uuid_lookup'); /** @var BlockContentUuidLookup $block_content_uuid_lookup */

  // Add link to edit block content in blocks layout page
  foreach (Element::children($form['blocks']) as $block_name) {
    if (isset($form['blocks'][$block_name]['operations'])) {
      $block_plugin = Block::load($block_name)->getPlugin(); /** @var BlockPluginInterface $block_plugin */

      if ($block_plugin->getBaseId() == 'block_content') {
        $block_content_id = $block_content_uuid_lookup->get($block_plugin->getDerivativeId());

        $form['blocks'][$block_name]['operations']['#links']['edit_content'] = [
          'title' => t('Edit content'),
          'url' => Url::fromRoute('entity.block_content.edit_form', ['block_content' => $block_content_id]),
          'weight' => 20,
        ];

        uasort($form['blocks'][$block_name]['operations']['#links'], '\Drupal\Component\Utility\SortArray::sortByWeightElement');
      }
    }
  }
}

/**
 * Implements hook_form_FORM_ID_alter(): block_content_form.
 */
function improvements_form_block_content_form_alter(array &$form, FormStateInterface $form_state) {
  // Add autofocus to content-block name field
  $form['info']['widget'][0]['value']['#attributes']['autofocus'] = TRUE;
}
