<?php

use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Menu\MenuLinkInterface;
use Drupal\Core\Render\Element;
use Drupal\Core\Url;
use Drupal\menu_link_content\Form\MenuLinkContentForm;
use Drupal\menu_link_content\MenuLinkContentInterface;
use Drupal\menu_ui\Form\MenuLinkEditForm;
use Drupal\menu_ui\MenuForm;
use Drupal\system\Entity\Menu;
use Drupal\system\MenuInterface;

/**
 * Implements hook_form_FORM_ID_alter(): menu_add_form.
 */
function improvements_form_menu_add_form_alter(array &$form, FormStateInterface $form_state) {
  if (isset($form['label'])) {
    $form['label']['#attributes']['autofocus'] = TRUE;
  }
}

/**
 * Implements hook_form_FORM_ID_alter(): menu_edit_form.
 */
function improvements_form_menu_edit_form_alter(array &$form, FormStateInterface $form_state) {
  if (!empty($form['links'])) {
    $menu_form_object = $form_state->getFormObject(); /** @var MenuForm $menu_form_object */
    $menu = $menu_form_object->getEntity(); /** @var MenuInterface $menu */
    $menu_id = $menu->id();
    $add_child_item_title = t('Add child item');
    $add_child_item_url = Url::fromRoute('entity.menu.add_link_form', ['menu' => $menu_id]);

    // Add link "Add child item" for menu link operations
    foreach (Element::children($form['links']['links']) as $key) {
      $form['links']['links'][$key]['operations']['#links']['add_child'] = [
        'title' => $add_child_item_title,
        'url' => $add_child_item_url,
        'query' => [
          'parent' => $menu_id . ':' . $form['links']['links'][$key]['id']['#value'],
        ],
      ];
    }
  }
}

/**
 * Implements hook_form_FORM_ID_alter(): menu_link_content_form.
 *
 * @see \Drupal\menu_link_content\Form\MenuLinkContentForm::form()
 */
function improvements_form_menu_link_content_form_alter(array &$form, FormStateInterface $form_state) {
  $menu_link_form_object = $form_state->getFormObject(); /** @var MenuLinkContentForm $menu_link_form_object */
  $menu_link = $menu_link_form_object->getEntity(); /** @var MenuLinkContentInterface $menu_link */

  // Add autofocus to menu link title
  $form['title']['widget'][0]['value']['#attributes']['autofocus'] = 'autofocus';

  // Remove "required" flag from link uri
  $form['link']['widget'][0]['uri']['#required'] = FALSE;

  // Set menu link parent if exists $_GET['parent'] param
  if ($parent = \Drupal::request()->query->get('parent')) {
    $form['menu_parent']['#default_value'] = $parent;
  }

  // Clear "weight" field on new link
  if ($menu_link->isNew()) {
    $form['weight']['widget'][0]['value']['#default_value'] = '';
  }

  $form['#entity_builders'][] = 'improvements_menu_link_content_entity_builder';
  $form['actions']['submit']['#submit'][] = 'improvements_menu_link_form_submit';
  $form['#pre_render'][] = '\Drupal\improvements\ImprovementsTrustedCallbacks::menuLinkContentFormPreRender';

  improvements_menu_link_advanced_settings_form_alter($form, $form_state);
}

/**
 * Implements hook_form_FORM_ID_alter(): menu_link_edit.
 *
 * @see \Drupal\menu_ui\Form\MenuLinkEditForm::buildForm()
 */
function improvements_form_menu_link_edit_alter(array &$form, FormStateInterface $form_state) {
  improvements_menu_link_advanced_settings_form_alter($form, $form_state);
  $form['#submit'][] = 'improvements_menu_link_form_submit';
}

/**
 * Menu link form submit callback.
 */
function improvements_menu_link_form_submit(array &$form, FormStateInterface $form_state) {
  // Redirect to menu edit form
  if (!\Drupal::request()->query->has('destination')) {
    $menu_link = $form_state->get('menu_link'); /** @var MenuLinkContentInterface|MenuLinkInterface $menu_link */
    $form_state->setRedirect('entity.menu.edit_form', ['menu' => $menu_link->getMenuName()]);
  }
}

/**
 * Menu link entity builder.
 */
function improvements_menu_link_content_entity_builder(string $entity_type, MenuLinkContentInterface $menu_link, array &$form, FormStateInterface $form_state) {
  // Set default url
  if (!$menu_link->get('link')->uri) {
    $menu_link->get('link')->uri = 'route:<none>';
  }

  // Calculate empty weight
  if ($form_state->getValue('weight')[0]['value'] === '') {
    $query = \Drupal::database()->select('menu_link_content_data', 'l');
    $query->addExpression('MAX(weight)');
    $query->condition('l.menu_name', $menu_link->getMenuName());
    $max_weight = $query->execute()->fetchField();

    $menu_link->set('weight', $max_weight + 1);
  }
}

/**
 * Preprocess function for menu.html.twig.
 */
function improvements_preprocess_menu(array &$vars) {
  // Set menu_name variable if not exists
  if (!isset($vars['menu_name'])) {
    $first_menu_item = reset($vars['items']);
    if (isset($first_menu_item['original_link'])) {
      $first_menu_link = $first_menu_item['original_link'];
      $vars['menu_name'] = $first_menu_link->getMenuName();
    }
  }

  if (!empty($vars['menu_name']) && ($menu = Menu::load($vars['menu_name']))) {
    $menu_links_advanced_options = $menu->getThirdPartySetting('improvements', 'links_options', []);
    improvements_preprocess_menu_items($vars, $vars['items'], $menu_links_advanced_options);
  }
}

/**
 * Preprocess menu items.
 *
 * @see improvements_preprocess_menu()
 */
function improvements_preprocess_menu_items(array &$menu_vars, array &$menu_items, array $menu_links_advanced_options) {
  foreach ($menu_items as &$menu_item) {
    $menu_link = $menu_item['original_link']; /** @var MenuLinkContentInterface|MenuLinkInterface $menu_link */

    // Set below text
    $item_link_definition = $menu_link->getPluginDefinition();
    if (!empty($item_link_definition['below_text'])) {
      $menu_item['below_text'] = $item_link_definition['below_text'];
    }

    // Advanced settings
    improvements_menu_link_advanced_settings_preprocess_menu_item($menu_item, $menu_vars, $menu_links_advanced_options);

    // Processing child items
    if ($menu_item['below']) {
      improvements_preprocess_menu_items($menu_vars, $menu_item['below'], $menu_links_advanced_options);
    }
  }
}

/**
 * Preprocess function for menu--toolbar.html.twig.
 */
function improvements_preprocess_menu__toolbar(array &$vars) {
  // Remove description from admin menu links
  _improvements_unset_menu_items_description($vars['items']);
}

/**
 * Remove description from menu links.
 */
function _improvements_unset_menu_items_description(array &$items) {
  foreach ($items as $key => $item) {
    $url = $item['url']; /** @var Url $url */
    $url_attributes = $url->getOption('attributes');

    if (isset($url_attributes['title'])) {
      unset($url_attributes['title']);
      $url->setOption('attributes', $url_attributes);
    }

    if (isset($item['below'])) {
      _improvements_unset_menu_items_description($item['below']);
    }
  }
}
