<?php

use Drupal\Component\Utility\NestedArray;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Render\BubbleableMetadata;

/**
 * Implements hook_site_third_party_settings_info().
 */
function improvements_site_third_party_settings_info() {
  return [
    'mail_for_notifications' => [
      'title' => t('E-mail address for notifications'),
      'description' => t('E-mail address for notifications of new comments, contact messages and other.'),
      'form_element_type' => 'textfield',
      'weight' => 1,
    ],
  ];
}

/**
 * Return site third-party settings info.
 */
function improvements_get_site_third_party_settings_info() {
  static $cache;

  if ($cache === NULL) {
    $cache = \Drupal::moduleHandler()->invokeAll('site_third_party_settings_info');
  }

  return $cache;
}

/**
 * Return site third-party settings.
 *
 * @return array|mixed
 */
function improvements_get_site_third_party_setting($setting_name = NULL) {
  static $site_third_party_settings;

  if ($site_third_party_settings === NULL) {
    $site_third_party_settings = \Drupal::config('improvements.site_third_party_settings')->get('settings');
  }

  if ($setting_name) {
    return $site_third_party_settings[$setting_name] ?? NULL;
  }

  return $site_third_party_settings;
}

/**
 * Implements hook_form_FORM_ID_alter(): system_site_information_settings.
 */
function improvements_form_system_site_information_settings_alter(array &$form, FormStateInterface $form_state) {
  $site_third_party_settings_info = improvements_get_site_third_party_settings_info();
  $site_third_party_settings = improvements_get_site_third_party_setting();

  foreach ($site_third_party_settings_info as $setting_name => $setting_info) {
    $group = $setting_info['group'] ?? 'site_information';
    $form_element_type = $setting_info['form_element_type'] ?? 'textfield';

    if (!isset($form[$group])) {
      $form[$group] = [
        '#type' => 'details',
        '#open' => TRUE,
      ];

      if (isset($setting_info['group_info'])) {
        $form[$group]['#title'] = $setting_info['group_info']['title'];
      }
    }

    $form[$group][$setting_name] = [
      '#type' => $form_element_type,
      '#title' => $setting_info['title'],
      '#description' => $setting_info['description'] ?? '',
      '#default_value' => $site_third_party_settings[$setting_name] ?? NULL,
      '#weight' => $setting_info['weight'] ?? 0,
    ];
  }

  $form['#submit'][] = 'improvements_form_system_site_information_settings_submit';
}

/**
 * Custom submit callback for system_site_information_settings form.
 */
function improvements_form_system_site_information_settings_submit(array &$form, FormStateInterface $form_state) {
  $site_third_party_settings_editable_config = \Drupal::configFactory()->getEditable('improvements.site_third_party_settings');
  $site_third_party_settings_info = improvements_get_site_third_party_settings_info();

  foreach ($site_third_party_settings_info as $setting_name => $setting_info) {
    $site_third_party_settings_editable_config->set('settings.' . $setting_name, $form_state->getValue($setting_name));
  }

  $site_third_party_settings_editable_config->save();
}

/**
 * Return site information.
 *
 * @param string|array|null
 *
 * @return array|string
 */
function improvements_get_site_information($setting_key = NULL) {
  static $site_information;

  if ($site_information === NULL) {
    $site_information = \Drupal::config('system.site')->get();
  }

  if ($setting_key) {
    return is_array($setting_key) ? NestedArray::getValue($site_information, $setting_key) : $site_information[$setting_key];
  }

  return $site_information;
}

/**
 * Implements hook_filter_functions().
 */
function improvements_filter_functions() {
  return [
    'site_information' => [
      'function' => 'function_filter__site_information',
    ],
  ];
}

/**
 * Filter function returning site information.
 */
function function_filter__site_information(string $setting_key) {
  if (strpos($setting_key, '.')) {
    $setting_key = explode('.', $setting_key);
  }

  return improvements_get_site_information($setting_key);
}

/**
 * @see improvements_token_info()
 */
function improvements_site_information_token_info() {
  $info = [];

  $site_third_party_settings_info = improvements_get_site_third_party_settings_info();
  foreach ($site_third_party_settings_info as $setting_name => $setting_info) {
    $info['tokens']['site'][$setting_name] = [
      'name' => $setting_info['title'],
      'description' => $setting_info['description'] ?? '',
    ];
  }

  return $info;
}

/**
 * @see improvements_tokens()
 */
function improvements_site_information_tokens(string $type, array $tokens, array $data, array $options, BubbleableMetadata $bubbleable_metadata) {
  $replacements = [];

  // [site:*]
  if ($type == 'site') {
    $site_third_party_settings = improvements_get_site_third_party_setting();

    foreach ($tokens as $name => $original) {
      if (!empty($site_third_party_settings[$name])) {
        $replacements[$original] = $site_third_party_settings[$name];
      }
    }
  }

  return $replacements;
}
