<?php

namespace Drupal\Tests\improvements\Functional;

use Drupal\comment\Tests\CommentTestTrait;
use Drupal\contact\Entity\ContactForm;
use Drupal\Core\Condition\ConditionInterface;
use Drupal\Core\Condition\ConditionManager;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Routing\RouteMatch;
use Drupal\Core\Test\AssertMailTrait;
use Drupal\Core\Url;
use Drupal\improvements\Cache\CookiesHashCacheContext;
use Drupal\improvements\Cache\EntityFieldCacheContext;
use Drupal\improvements\Cache\EntityTypeCacheContext;
use Drupal\improvements\Cache\IsEntityPageCacheContext;
use Drupal\improvements\Cache\ParentTermCacheContext;
use Drupal\improvements\ImprovementsHelper;
use Drupal\menu_link_content\Entity\MenuLinkContent;
use Drupal\node\Entity\Node;
use Drupal\pathauto\Entity\PathautoPattern;
use Drupal\taxonomy\Entity\Term;
use Drupal\taxonomy\Entity\Vocabulary;
use Drupal\Tests\block\Traits\BlockCreationTrait;
use Drupal\Tests\BrowserTestBase;
use Drupal\Tests\improvements\Traits\ImprovementsTestTrait;
use Drupal\Tests\node\Traits\NodeCreationTrait;
use Drupal\Tests\taxonomy\Traits\TaxonomyTestTrait;
use Drupal\user\Entity\Role;
use Drupal\user\RoleInterface;
use Drupal\views\ViewExecutable;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\RequestStack;
use Symfony\Component\Routing\Route;

class ImprovementsTest extends BrowserTestBase {

  use AssertMailTrait;
  use CommentTestTrait;
  use BlockCreationTrait;
  use ImprovementsTestTrait;
  use NodeCreationTrait;
  use TaxonomyTestTrait;

  /**
   * {@inheritDoc}
   */
  protected $defaultTheme = 'classy';

  /**
   * {@inheritDoc}
   */
  protected static $modules = [
    'improvements',
    'improvements_test',
    'block',
    'block_content',
    'breakpoint',
    'comment',
    'contact',
    'datetime',
    'field_ui',
    'file',
    'filter',
    'image',
    'menu_link_content',
    'menu_ui',
    'node',
    'taxonomy',
    'path',
    'pathauto',
    'user',
    'views',
    'views_ui',
  ];

  /**
   * @var \Drupal\Core\Render\RendererInterface
   */
  protected $renderer;

  /**
   * @var \Drupal\Core\Entity\EntityDisplayRepositoryInterface
   */
  protected $entityDisplayRepository;

  /**
   * @var \Drupal\Core\Config\ConfigFactoryInterface
   */
  protected $configFactory;

  /**
   * @var \Drupal\token\TokenInterface
   */
  protected $tokenService;

  /**
   * @var \Drupal\node\NodeTypeInterface
   */
  protected $nodeType;

  /**
   * @var \Drupal\taxonomy\VocabularyInterface
   */
  protected $vocabulary;

  /**
   * {@inheritDoc}
   */
  protected function setUp() {
    parent::setUp();

    $this->renderer = $this->container->get('renderer');
    $this->entityDisplayRepository = $this->container->get('entity_display.repository');
    $this->configFactory = $this->container->get('config.factory');
    $this->tokenService = $this->container->get('token');

    // Create "page" node type
    $this->nodeType = $this->createContentType(['type' => 'page', 'name' => 'Page']);
    $this->addDefaultCommentField('node', 'page');

    // Create "category" vocabulary
    $this->vocabulary = Vocabulary::create(['name' => 'Categories', 'vid' => 'category']);
    $this->vocabulary->save();

    // Create content block type
    $this->createBlockContentType();

    // Place blocks
    $this->placeBlock('system_menu_block:main', [
      'id' => 'main_menu_block',
      'region' => 'help',
      'weight' => 0,
      'label' => 'Main menu',
    ]);
    $this->placeBlock('page_title_block', [
      'id' => 'page_title_block',
      'region' => 'help',
      'weight' => 1,
    ]);
    $this->placeBlock('local_tasks_block', [
      'id' => 'local_tasks_block',
      'region' => 'help',
      'weight' => 2,
    ]);
    $this->placeBlock('local_actions_block', [
      'id' => 'local_actions_block',
      'region' => 'help',
      'weight' => 3,
    ]);

    // Create filter formats
    $this->createFilterFormat(['format' => 'full_html', 'name' => 'Full HTML']);
    $this->createFilterFormat(['format' => 'raw_html', 'name' => 'Raw HTML']);

    // Setting-up user permissions
    $anonymous_role = Role::load(Role::ANONYMOUS_ID); /* @var RoleInterface $anonymous_role */
    $anonymous_role->grantPermission('access comments');
    $anonymous_role->grantPermission('access site-wide contact form');
    $anonymous_role->save();

    $this->configFactory->getEditable('views.settings')->set('display_extenders', ['improvements_display_extender'])->save();
  }

  /**
   * Main test function.
   */
  public function testMain() {
    // Run all test with "_test" prefix
    $class_methods = get_class_methods(self::class);
    foreach ($class_methods as $class_method) {
      if (strpos($class_method, '_test') === 0) {
        $result = $this->{$class_method}();
        if ($result === FALSE) {
          break;
        }
      }
    }
  }

  /**
   * Test "option" element attributes.
   */
  private function _testFormSelectOptionsAttributes() {
    $this->drupalGet('/improvements-test/test-form');
    $this->dontSeeErrorMessage();
    $web_assert = $this->assertSession();
    $web_assert->elementExists('css', 'select[name="select1"] option[value="foo"]');
    $web_assert->elementExists('css', 'select[name="select2"] option[value="foo"][data-custom-attribute="foo-custom-attribute"].foo-test-class');
    $web_assert->elementExists('css', 'select[name="select2"] option[value="bar"]:disabled');
    $web_assert->elementExists('css', 'select[name="select2"] option[value="baz"]');

    $this->drupalPostForm(NULL, [
      'select1' => 'bar',
      'select2' => 'baz',
    ], 'Submit');
    $this->dontSeeErrorMessage();
    $web_assert->pageTextContains('[select1] => bar');
    $web_assert->pageTextContains('[select2] => baz');
  }

  /**
   * Test "managed_file" element improvements.
   */
  private function _testFormManagedFileImprovements() {
    $this->drupalGet('/improvements-test/test-form');
    $this->dontSeeErrorMessage();
    $web_assert = $this->assertSession();
    $web_assert->elementExists('css', '#edit-file-without-autoupload-upload.js-form-file--no-autoupload');
    $web_assert->elementExists('css', '#edit-file-without-autoupload-upload-button.js-hide');
    $web_assert->elementExists('css', '#edit-file-with-button-upload-button');
    $web_assert->elementNotExists('css', '#edit-file-with-button-upload-button.js-hide');
    $web_assert->elementNotExists('css', '#edit-file-without-button-upload-button');
  }

  /**
   * Test '#title_display'=>'placeholder'.
   */
  private function _testFormInputTitleDisplayPlaceholder() {
    $this->drupalGet('/improvements-test/test-form');
    $this->dontSeeErrorMessage();
    $web_assert = $this->assertSession();
    $web_assert->elementAttributeContains('css', '#edit-label-in-placeholder', 'placeholder', 'Label in placeholder');

    $this->drupalPostForm(NULL, [
      'label_in_placeholder' => 'foo',
    ], 'Submit');
    $this->dontSeeErrorMessage();
    $web_assert->pageTextContains('[label_in_placeholder] => foo');
  }

  /**
   * Test '#type'=>'range_slider'.
   *
   * @covers \Drupal\improvements\Element\RangeSlider
   */
  private function _testFormElementRangeSlider() {
    $this->drupalGet('/improvements-test/test-form');
    $this->dontSeeErrorMessage();
    $this->drupalPostForm(NULL, [
      'range_slider[from]' => 5,
      'range_slider[to]' => 60,
    ], 'Submit');
    $this->dontSeeErrorMessage();
    $web_assert = $this->assertSession();
    $web_assert->pageTextContains('[from] => 5');
    $web_assert->pageTextContains('[to] => 60');
  }

  /**
   * Test tracking form submit functionality.
   */
  private function _testFormTrackingSubmit() {
    $this->drupalPostForm('/improvements-test/test-form', [], 'Submit');
    $this->dontSeeErrorMessage();
    $drupal_settings = $this->getDrupalSettings();
    $this->assertIdentical('improvements_test_form', $drupal_settings['submittedForms'][0]);
  }

  /**
   * Test '#theme'=>'simple_field'.
   */
  private function _testThemeSimpleField() {
    $build = [
      '#theme' => 'simple_field',
      '#field_name' => 'custom_field',
      '#title' => 'Custom field',
      '#entity_type' => 'node',
      '#bundle' => 'article',
      '#items' => 'Field value',
    ];
    $html = (string)$this->renderer->renderRoot($build);
    $this->assertStringContainsString($build['#title'], $html);
    $this->assertStringContainsString($build['#items'], $html);
  }

  /**
   * Test '#theme'=>'picture'.
   */
  private function _testThemePicture() {
    $build = [
      '#theme' => 'picture',
      '#img' => [
        'src' => '/sites/default/files/logo.jpg',
        'alt' => '',
      ],
      '#source' => [
        'mobile' => [
          'srcset' => '/sites/default/files/logo-mobile.jpg',
          'media' => '(min-width: 600px)',
        ],
      ],
    ];
    $html = (string)$this->renderer->renderRoot($build);
    $this->assertStringContainsString('<picture>', $html);
    $this->assertStringContainsString('<source srcset="/sites/default/files/logo-mobile.jpg" media="(min-width: 600px)" />', $html);
    $this->assertStringContainsString('<img src="/sites/default/files/logo.jpg" alt="" />', $html);
  }

  /**
   * Test '#theme_wrappers'=>['noindex'].
   */
  private function _testThemeNoindex() {
    $build = [
      '#markup' => 'Markup',
      '#theme_wrappers' => ['noindex'],
    ];
    $html = (string)$this->renderer->renderRoot($build);
    $this->assertStringContainsString("<!--noindex-->\nMarkup\n<!--/noindex-->", $html);
  }

  /**
   * Test altering page title.
   *
   * @covers \Drupal\improvements\ImprovementsTrustedCallbacks
   */
  private function _testPageTitleAlter() {
    $this->drupalGet('/improvements-test/altered-route-title');
    $this->dontSeeErrorMessage();
    $this->assertSession()->elementTextContains('css', 'h1', 'Altered route title');
  }

  /**
   * Test remove "off canvas page wrapper" for anonymouse users.
   *
   * @covers \Drupal\improvements\ImprovementsTrustedCallbacks::pagePreRender
   */
  private function _testRemoveOffCanvasPageWrapper() {
    $this->drupalGet('<front>');
    $this->dontSeeErrorMessage();
    $this->assertSession()->elementNotExists('css', '.dialog-off-canvas-main-canvas');

    $this->drupalLoginAsRoot();
    $this->drupalGet('<front>');
    $this->dontSeeErrorMessage();
    $this->assertSession()->elementExists('css', '.dialog-off-canvas-main-canvas');
    $this->drupalLogout();
  }

  /**
   * Test create link with attribute href="#".
   *
   * @covers \Drupal\improvements\ImprovementsTrustedCallbacks::linkPostRender
   */
  private function _testLinkWithOnlySharpSymbol() {
    $build = [
      '#type' => 'link',
      '#url' => Url::fromRoute('<none>', [], ['fragment' => '<none>']),
      '#title' => 'Test',
    ];
    $html = (string)$this->renderer->renderRoot($build);
    $this->assertIdentical('<a href="#">' . $build['#title'] . '</a>', $html);
  }

  /**
   * Test js theme path settings.
   */
  private function _testThemePathsInDrupalSettings() {
    $this->drupalGet('<front>');
    $drupal_settings = $this->getDrupalSettings();
    $this->assertIdentical('core/themes/' . $this->defaultTheme, $drupal_settings['path']['currentThemePath']);
    $this->assertIdentical('core/themes/' . $this->defaultTheme, $drupal_settings['path']['defaultThemePath']);
  }

  /**
   * Test modules list filter autofocus.
   *
   * @covers ::improvements_form_system_modules_alter
   */
  private function _testModulesListAutofocus() {
    $this->drupalLoginAsRoot();
    $this->drupalGet('/admin/modules');
    $this->dontSeeErrorMessage();
    $this->assertSession()->elementExists('css', '#edit-text[autofocus]');
    $this->drupalLogout();
  }

  /**
   * Test ComputedMapFieldItemList class.
   *
   * @covers \Drupal\improvements\Plugin\Field\ComputedMapFieldItemList
   */
  private function _testFieldTypeComputedMap() {
    $node = $this->createNode([
      'title' => 'Node for test ' . __FUNCTION__,
      'test_map_field' => [
        'first' => 'First',
        'second' => [
          'foo' => 'Foo',
          'bar' => 'Bar',
        ],
      ],
    ]);

    $this->assertIdentical('Foo', $node->get('test_computed_map_field_single')->value);
    $this->assertIdentical('Foo-Bar', $node->get('test_computed_map_field_multiple')->value);
    $this->assertIdentical('Foo', $node->get('test_computed_map_field_single')->first()->value);
    $this->assertIdentical('Foo-Bar', $node->get('test_computed_map_field_multiple')->first()->value);
    $this->assertIdentical([0 => ['value' => 'Foo']], $node->get('test_computed_map_field_single')->getValue());
    $this->assertIdentical([0 => ['value' => 'Foo-Bar']], $node->get('test_computed_map_field_multiple')->getValue());
    $this->assertIdentical(['value' => 'Foo'], $node->get('test_computed_map_field_single')->first()->getValue());
    $this->assertIdentical(['value' => 'Foo-Bar'], $node->get('test_computed_map_field_multiple')->first()->getValue());

    $node->delete();
  }

  /**
   * Test ThirdPartyDataItem.
   *
   * @covers \Drupal\improvements\Plugin\Field\FieldType\ThirdPartyDataItem
   */
  private function _testFieldTypeThirdPartyData() {
    $node = $this->createNode(['title' => 'Node for test ' . __FUNCTION__]);
    $node->get('test_third_party_data')->setThirdPartyData('improvements_test', 'foo', 'bar');
    $node->get('test_third_party_data')->setThirdPartyData('improvements_test', 'baz', 'qux');
    $node->save();
    $node = Node::load($node->id());

    $this->assertIdentical('bar', $node->get('test_third_party_data')->getThirdPartyData('improvements_test', 'foo'));
    $this->assertIdentical('qux', $node->get('test_third_party_data')->getThirdPartyData('improvements_test', 'baz'));
    $this->assertIdentical(NULL, $node->get('test_third_party_data')->getThirdPartyData('improvements_test', 'bat'));

    $node->delete();
  }

  /**
   * Test textarea widget improvements.
   */
  private function _testFieldWidgetTextareaImprovements() {
    $web_assert = $this->assertSession();
    $this->drupalLoginAsRoot();
    $this->createField('node', 'page', 'field_text', 'text_long', 'Text', [], [], 'text_textarea');

    $this->drupalGet('/node/add/page');
    $this->dontSeeErrorMessage();
    $web_assert->elementExists('css', 'select[name="field_text[0][format]"]');

    $this->drupalGet('/admin/structure/types/manage/page/form-display');
    $this->openEntityDisplayComponentSettings('field_text');
    $web_assert->fieldExists('fields[field_text][settings_edit_form][third_party_settings][improvements][max_length]');
    $web_assert->fieldExists('fields[field_text][settings_edit_form][third_party_settings][improvements][dynamic_height]');
    $web_assert->fieldExists('fields[field_text][settings_edit_form][third_party_settings][improvements][disable_editor]');
    $web_assert->fieldExists('fields[field_text][settings_edit_form][third_party_settings][improvements][hide_format_select]');

    $this->drupalPostForm(NULL, [
      'fields[field_text][settings_edit_form][third_party_settings][improvements][max_length]' => 10,
      'fields[field_text][settings_edit_form][third_party_settings][improvements][dynamic_height]' => TRUE,
      'fields[field_text][settings_edit_form][third_party_settings][improvements][disable_editor]' => TRUE,
      'fields[field_text][settings_edit_form][third_party_settings][improvements][hide_format_select]' => TRUE,
    ], 'Save');
    $this->dontSeeErrorMessage();

    $this->drupalGet('/node/add/page');
    $this->dontSeeErrorMessage();
    $web_assert = $this->assertSession();
    $web_assert->elementAttributeContains('css', '#edit-field-text-0-value', 'maxlength', '10');
    $web_assert->elementNotExists('css', 'select[name="field_text[0][format]"]');

    // @TODO Add test disabled CKEditor

    $this->deleteField('field_text');
    $this->drupalLogout();
  }

  /**
   * Test textfield widget improvements.
   */
  private function _testFieldWidgetTextfieldImprovements() {
    $this->createField('node', 'page', 'field_string', 'string', 'String', [], [], 'string_textfield');
    $this->drupalLoginAsRoot();
    $web_assert = $this->assertSession();

    $this->drupalGet('/admin/structure/types/manage/page/form-display');
    $this->openEntityDisplayComponentSettings('field_string');
    $web_assert->fieldExists('fields[field_string][settings_edit_form][third_party_settings][improvements][max_length]');

    $this->drupalPostForm(NULL, [
      'fields[field_string][settings_edit_form][third_party_settings][improvements][max_length]' => 10,
    ], 'Save');
    $this->dontSeeErrorMessage();

    $this->drupalGet('/node/add/page');
    $web_assert->elementAttributeContains('css', '#edit-field-string-0-value', 'maxlength', '10');

    $this->deleteField('field_string');
    $this->drupalLogout();
  }

  /**
   * Test "file" widget improvements.
   */
  private function _testFieldWidgetFileImprovements() {
    $this->createField('node', 'page', 'field_file', 'file', 'File', [], [], 'file_generic');
    $this->drupalLoginAsRoot();
    $web_assert = $this->assertSession();

    $this->drupalGet('/node/add/page');
    $web_assert->elementExists('css', '#edit-field-file-0-upload');

    $this->drupalGet('/admin/structure/types/manage/page/form-display');
    $this->openEntityDisplayComponentSettings('field_file');
    $web_assert->fieldExists('fields[field_file][settings_edit_form][third_party_settings][improvements][autoupload]');
    $web_assert->fieldExists('fields[field_file][settings_edit_form][third_party_settings][improvements][upload_button_state]');

    $this->drupalPostForm(NULL, [
      'fields[field_file][settings_edit_form][third_party_settings][improvements][autoupload]' => FALSE,
    ], 'Save');
    $this->dontSeeErrorMessage();

    $this->drupalGet('/node/add/page');
    $web_assert->elementExists('css', '#edit-field-file-0-upload.js-form-file--no-autoupload');
    $web_assert->elementExists('css', '#edit-field-file-0-upload-button.js-hide');

    $field_file_widget_settings['third_party_settings']['improvements']['upload_button_state'] = 1;
    $this->setFieldWidgetSettings('node', 'page', 'field_file', $field_file_widget_settings);
    $this->drupalGet('/node/add/page');
    $web_assert->elementExists('css', '#edit-field-file-0-upload-button');
    $web_assert->elementNotExists('css', '#edit-field-file-0-upload-button.js-hide');

    $field_file_widget_settings['third_party_settings']['improvements']['upload_button_state'] = 0;
    $this->setFieldWidgetSettings('node', 'page', 'field_file', $field_file_widget_settings);
    $this->drupalGet('/node/add/page');
    $web_assert->elementNotExists('css', '#edit-field-file-0-upload-button');

    $this->deleteField('field_file');
    $this->drupalLogout();
  }

  /**
   * Test "date" widget functionality.
   */
  private function _testFieldWidgetDateImprovements() {
    $this->createField('node', 'page', 'field_date', 'datetime', 'Date', [], [], 'datetime_default');
    $this->drupalLoginAsRoot();
    $this->drupalGet('/node/add/page');
    $this->dontSeeErrorMessage();

    $web_assert = $this->assertSession();
    $web_assert->elementExists('css', '.field--name-field-date');
    $web_assert->elementExists('css', '#edit-field-date-0-value-date');
    $web_assert->elementNotExists('css', '.field--name-field-date fieldset');

    $this->deleteField('field_date');
    $this->drupalLogout();
  }

  /**
   * Test field widget "multiple_string_textarea".
   *
   * @covers \Drupal\improvements\Plugin\Field\FieldWidget\MultipleStringTextarea
   */
  private function _testFieldWidgetMultipleStringTextarea() {
    $this->createField('node', 'page', 'field_strings', 'string', 'Strings', ['cardinality' => -1], [], 'multiple_string_textarea');

    $this->drupalLoginAsRoot();
    $this->drupalGet('/node/add/page');
    $this->dontSeeErrorMessage();
    $web_assert = $this->assertSession();
    $web_assert->elementExists('css', 'textarea[name="field_strings[value]"]');

    $this->drupalPostForm(NULL, [
      'title[0][value]' => 'Node for test ' . __FUNCTION__,
      'field_strings[value]' => "foo\nbar\nbaz",
    ], 'Save');
    $this->dontSeeErrorMessage();
    $node_id = $this->getLastAddedNodeId();
    $node = Node::load($node_id);
    $this->assertIdentical([
      0 => ['value' => 'foo'],
      1 => ['value' => 'bar'],
      2 => ['value' => 'baz'],
    ], $node->get('field_strings')->getValue());

    $this->deleteEntities($node);
    $this->deleteField('field_strings');
    $this->drupalLogout();
  }

  /**
   * Test "entity_reference_views_formatter" formatter.
   *
   * @covers \Drupal\improvements\Plugin\Field\FieldFormatter\EntityReferenceViewsFormatter
   */
  private function _testFieldFormatterEntityReferenceViews() {
    $this->drupalLoginAsRoot();
    $web_assert = $this->assertSession();
    $field_name = 'field_views_formatter_test';
    $views_name = 'test_entity_reference_views_formatter';

    $this->createField('node', 'page', $field_name, 'entity_reference', 'Entity reference', ['cardinality' => -1], [], NULL, [], 'entity_reference_views_formatter', ['weight' => 0]);

    $this->drupalGet('/admin/structure/types/manage/page/display');
    $web_assert->elementExists('css', 'option[value="entity_reference_views_formatter"][selected]');

    $this->openEntityDisplayComponentSettings($field_name);
    $web_assert->elementExists('css', 'select[name="fields[' . $field_name . '][settings_edit_form][settings][views]"]');
    $web_assert->elementExists('css', 'option[value="' . $views_name . ':default"]');

    $this->drupalPostForm(NULL, [
      'fields[' . $field_name . '][settings_edit_form][settings][views]' => $views_name . ':default',
    ], 'Save');
    $this->dontSeeErrorMessage();
    $web_assert->pageTextContains('Views: ' . $views_name . ':default');

    $node1 = $this->createNode(['title' => 'Node 1 for test ' . __FUNCTION__]);
    $node2 = $this->createNode(['title' => 'Node 2 for test ' . __FUNCTION__]);
    $node3 = $this->createNode(['title' => 'Node 3 for test ' . __FUNCTION__]);
    $node4 = $this->createNode([
      'title' => 'Node 4 for test ' . __FUNCTION__,
      $field_name => [$node1->id(), $node2->id()],
    ]);

    $this->drupalGetEntityPage($node4);
    $web_assert->elementExists('css', '.view-' . str_replace('_', '-', $views_name));
    $web_assert->pageTextContains($node1->label());
    $web_assert->pageTextContains($node2->label());
    $web_assert->pageTextNotContains($node3->label());

    $this->drupalGetEntityPage($node3);
    $web_assert->elementNotExists('css', '.view-' . str_replace('_', '-', $views_name));

    // Clean environment
    $this->deleteEntities($node1, $node2, $node3, $node4);
    $this->deleteField($field_name);
    $this->drupalLogout();
  }

  /**
   * Test ExtendedImageFormatter
   *
   * @covers \Drupal\improvements\Plugin\Field\FieldFormatter\ExtendedImageFormatter
   */
  private function _testFieldFormatterExtendedImage() {
    $web_assert = $this->assertSession();

    $this->createField('node', 'page', 'field_image', 'image', 'Image', ['cardinality' => -1], [], 'image_image', [], 'extended_image_formatter');
    $this->drupalLoginAsRoot();
    $this->drupalGet('/admin/structure/types/manage/page/display');
    $this->dontSeeErrorMessage();
    $this->openEntityDisplayComponentSettings('field_image');
    $web_assert->elementTextContains('css', '.plugin-name', 'Extended image formatter');
    $web_assert->fieldExists('fields[field_image][settings_edit_form][settings][image_style]');
    $web_assert->fieldExists('fields[field_image][settings_edit_form][settings][first_image_style]');
    $web_assert->fieldExists('fields[field_image][settings_edit_form][settings][image_link]');
    $web_assert->fieldExists('fields[field_image][settings_edit_form][settings][image_link_style]');

    $this->drupalPostForm(NULL, [
      'fields[field_image][settings_edit_form][settings][image_style]' => 'thumbnail',
      'fields[field_image][settings_edit_form][settings][first_image_style]' => 'large',
      'fields[field_image][settings_edit_form][settings][image_link]' => 'file',
      'fields[field_image][settings_edit_form][settings][image_link_style]' => 'medium',
    ], 'Save');
    $this->dontSeeErrorMessage();
    $web_assert->pageTextContains('Image style: thumbnail');

    $first_image = $this->createFileEntity(['uri' => 'public://first-image.png']);
    $second_image = $this->createFileEntity(['uri' => 'public://second-image.png']);
    $node = $this->createNode([
      'title' => 'Node for test ' . __FUNCTION__,
      'field_image' => [
        $first_image,
        $second_image
      ],
    ]);
    $this->drupalGetEntityPage($node);
    $web_assert->elementExists('css', '.field--name-field-image .field__item:nth-child(1) .image-style-large');
    $web_assert->elementExists('css', '.field--name-field-image .field__item:nth-child(2) .image-style-thumbnail');

    // Clean environment
    $node->delete();
    $first_image->delete();
    $second_image->delete();
    $this->deleteField('field_image');
    $this->drupalLogout();
  }

  /**
   * Test formatter "image_with_mobile_alternative".
   *
   * @covers \Drupal\improvements\Plugin\Field\FieldFormatter\ImageWithMobileAlternativeFormatter
   */
  private function _testFieldFormatterImageWithMobileAlternative() {
    $web_assert = $this->assertSession();

    // Preparing environment
    $this->drupalLoginAsRoot();
    $this->createField('node', 'page', 'field_image', 'image', 'Image', [], [], 'image_image', [], 'image_with_mobile_alternative', [
      'settings' => [
        'breakpoint_group' => 'improvements_test',
      ]
    ]);
    $this->createField('node', 'page', 'field_mobile_image', 'image', 'Mobile image');

    $this->drupalGet('/admin/structure/types/manage/page/display');
    $this->dontSeeErrorMessage();
    $this->openEntityDisplayComponentSettings('field_image');
    $web_assert->elementTextContains('css', '.plugin-name', 'Image with mobile alternative');
    $web_assert->fieldExists('fields[field_image][settings_edit_form][settings][image_style]');
    $web_assert->fieldExists('fields[field_image][settings_edit_form][settings][breakpoint_group]');
    $web_assert->fieldExists('fields[field_image][settings_edit_form][settings][mobile_breakpoint]');
    $web_assert->fieldExists('fields[field_image][settings_edit_form][settings][mobile_field]');
    $web_assert->fieldExists('fields[field_image][settings_edit_form][settings][mobile_image_style]');

    $this->drupalPostForm(NULL, [
      'fields[field_image][settings_edit_form][settings][mobile_breakpoint]' => 'improvements_test.mobile',
      'fields[field_image][settings_edit_form][settings][mobile_field]' => 'field_mobile_image',
    ], 'Save');
    $this->dontSeeErrorMessage();
    $web_assert->pageTextContains('Breakpoint group: improvements_test');
    $web_assert->pageTextContains('Mobile breakpoint: improvements_test.mobile');
    $web_assert->pageTextContains('Mobile field: field_mobile_image');

    $first_image = $this->createFileEntity(['uri' => 'public://first-image.png']);
    $second_image = $this->createFileEntity(['uri' => 'public://second-image.png']);
    $node = $this->createNode([
      'title' => 'Node for test ' . __FUNCTION__,
      'field_image' => $first_image,
      'field_mobile_image' => $second_image,
    ]);
    $this->drupalGetEntityPage($node);
    $web_assert->elementExists('css', '.field--name-field-image picture');
    $web_assert->elementExists('css', '.field--name-field-image source[srcset="' . $second_image->createFileUrl() . '"][type="image/png"][media="(max-width: 640px)"]');
    $web_assert->elementExists('css', '.field--name-field-image img[src="' . $first_image->createFileUrl() . '"]');

    // Clean environment
    $this->deleteEntities($node, $first_image, $second_image);
    $this->deleteField('field_image');
    $this->deleteField('field_mobile_image');
    $this->drupalLogout();
  }

  /**
   * Test formatter "number_raw".
   *
   * @covers \Drupal\improvements\Plugin\Field\FieldFormatter\NumericRawFormatter
   */
  private function _testFieldFormatterNumberRaw() {
    $this->createField('node', 'page', 'field_number', 'decimal', 'Number', [], [], NULL, [], 'number_raw');
    $this->drupalLoginAsRoot();
    $this->drupalGet('/admin/structure/types/manage/page/display');
    $this->dontSeeErrorMessage();
    $web_assert = $this->assertSession();
    $web_assert->elementExists('css', 'select[name="fields[field_number][type]"] option[value="number_raw"][selected]');

    $this->openEntityDisplayComponentSettings('field_number');
    $web_assert->elementTextContains('css', '.plugin-name', 'Raw');
    $web_assert->fieldExists('fields[field_number][settings_edit_form][settings][thousand_separator]');
    $web_assert->fieldExists('fields[field_number][settings_edit_form][settings][prefix_suffix]');

    $node = $this->createNode([
      'title' => 'Node for test ' . __FUNCTION__,
      'field_number' => 0.10,
    ]);
    $this->drupalGetEntityPage($node);
    $web_assert->elementTextContains('css', '.field--name-field-number .field__item', '0.10');

    $this->deleteEntities($node);
    $this->deleteField('field_number');
  }

  /**
   * Test field formatter "string_list_formatter".
   *
   * @covers \Drupal\improvements\Plugin\Field\FieldFormatter\StringListFormatter
   */
  private function _testFieldFormatterStringList() {
    $this->createField('node', 'page', 'field_strings', 'string', 'Strings', ['cardinality' => -1], [], NULL, [], 'string_list_formatter');
    $this->drupalLoginAsRoot();
    $this->drupalGet('/admin/structure/types/manage/page/display');
    $this->dontSeeErrorMessage();
    $web_assert = $this->assertSession();
    $web_assert->elementExists('css', 'select[name="fields[field_strings][type]"] option[value="string_list_formatter"][selected]');

    $this->openEntityDisplayComponentSettings('field_strings');
    $web_assert->fieldExists('fields[field_strings][settings_edit_form][settings][list_class]');

    $list_class = 'list-test-class';
    $this->drupalPostForm(NULL, [
      'fields[field_strings][settings_edit_form][settings][list_class]' => $list_class,
    ], 'Save');
    $this->dontSeeErrorMessage();
    $web_assert->pageTextContains('List html class: ' . $list_class);

    $node = $this->createNode([
      'title' => 'Node for test ' . __FUNCTION__,
      'field_strings' => [
        'foo',
        'bar',
      ],
    ]);
    $this->drupalGetEntityPage($node);
    $this->dontSeeErrorMessage();
    $web_assert->elementTextContains('css', 'ul.' . $list_class . ' li:nth-child(1)', 'foo');
    $web_assert->elementTextContains('css', 'ul.' . $list_class . ' li:nth-child(2)', 'bar');

    $this->deleteEntities($node);
    $this->deleteField('field_strings');
    $this->drupalLogout();
  }

  /**
   * Test field formatter "text_without_escaping".
   *
   * @covers \Drupal\improvements\Plugin\Field\FieldFormatter\TextWithoutEscapingFormatter
   */
  private function _testFieldFormatterTextWithoutEscaping() {
    $this->createField('node', 'page', 'field_string', 'string', 'String', [], [], NULL, [], 'text_without_escaping');
    $this->drupalLoginAsRoot();
    $this->drupalGet('/admin/structure/types/manage/page/display');
    $this->dontSeeErrorMessage();
    $web_assert = $this->assertSession();
    $web_assert->elementExists('css', 'select[name="fields[field_string][type]"] option[value="text_without_escaping"][selected]');

    $node = $this->createNode([
      'title' => 'Node for test ' . __FUNCTION__,
      'field_string' => 'foo <b class="test-bar-class">bar</b> baz',
    ]);
    $this->drupalGetEntityPage($node);
    $this->dontSeeErrorMessage();
    $web_assert->elementExists('css', '.test-bar-class');

    $this->deleteEntities($node);
    $this->deleteField('field_string');
    $this->drupalLogout();
  }

  /**
   * Test link for edit block content.
   *
   * @covers ::improvements_form_block_admin_display_form_alter
   */
  private function _testContentBlockEditLink() {
    $block_content = $this->createBlockContent();
    $block = $this->placeBlockContent($block_content, ['id' => 'test_edit_content_link']);
    $this->drupalLoginAsRoot();
    $this->drupalGet('/admin/structure/block');
    $this->dontSeeErrorMessage();
    $this->assertSession()->elementExists('css', 'tr[data-drupal-selector="edit-blocks-test-edit-content-link"] .dropbutton a[href="/block/' . $block_content->id() . '"]');

    $block->delete();
    $block_content->delete();
    $this->drupalLogout();
  }

  /**
   * Test content block form autofocus.
   *
   * @covers ::improvements_form_block_content_form_alter
   */
  private function _testBlockContentAutofocus() {
    $this->drupalLoginAsRoot();
    $this->drupalGet('/block/add/basic');
    $this->dontSeeErrorMessage();
    $this->assertSession()->elementAttributeExists('css', '#edit-info-0-value', 'autofocus');
    $this->drupalLogout();
  }

  /**
   * Test remove commented entity on comment form.
   *
   * @covers ::improvements_page_comment_reply_result_alter
   */
  private function _testRemoveCommentedEntityOnCommentForm() {
    $node = $this->createNode(['title' => 'Node for test ' . __FUNCTION__]);
    $this->drupalLoginAsRoot();
    $this->drupalGet('/comment/reply/node/' . $node->id() . '/comment');
    $this->dontSeeErrorMessage();

    $web_assert = $this->assertSession();
    $web_assert->elementExists('css', '.comment-comment-form');
    $web_assert->elementNotExists('css', 'article');

    $node->delete();
    $this->drupalLogout();
  }

  /**
   * Test comment display fields "uid" and "created".
   *
   * @covers ::improvements_entity_base_field_info_alter
   */
  private function _testCommentShowAuthorAndDateFields() {
    $this->setFieldFormatterSettings('comment', 'comment', 'uid', ['type' => 'author']);
    $this->setFieldFormatterSettings('comment', 'comment', 'created', ['type' => 'timestamp']);

    $node = $this->createNode([
      'title' => 'Node for test ' . __FUNCTION__,
      'body' => 'Test node content',
    ]);
    $this->createComment($node);

    $this->drupalLoginAsRoot();
    $this->drupalGetEntityPage($node);
    $this->dontSeeErrorMessage();

    $web_assert = $this->assertSession();
    $web_assert->elementExists('css', '.comment .field--name-uid');
    $web_assert->elementExists('css', '.comment .field--name-created');

    $node->delete();
    $this->drupalLogout();
  }

  /**
   * Test simple contact form.
   */
  private function _testContactFormWithoutAdditionalSettings() {
    $this->drupalLoginAsRoot();

    $this->drupalGet('/admin/structure/contact/add');
    $this->dontSeeErrorMessage();
    $web_assert = $this->assertSession();

    $contact_form_values = [
      'label' => 'Contact form without advanced settings',
      'id' => 'simple_contact_form',
      'recipients' => 'test@example.com',
      'message' => 'Form submitted',
    ];
    $this->drupalPostForm('/admin/structure/contact/add', $contact_form_values, 'Save');
    $this->dontSeeErrorMessage();
    $web_assert->pageTextContains('has been added');

    $this->drupalGet('/contact/' . $contact_form_values['id']);
    $this->dontSeeErrorMessage();
    $web_assert->elementTextContains('css', '.page-title', $contact_form_values['label']);
    $web_assert->elementTextContains('css', '#edit-name label', 'Your name');
    $web_assert->elementTextContains('css', '#edit-mail label', 'Your email address');
    $web_assert->elementTextContains('css', '#edit-subject-wrapper label', 'Subject');
    $web_assert->elementTextContains('css', '#edit-message-wrapper label', 'Message');
    $web_assert->elementAttributeExists('css', '#edit-message-0-value', 'required');

    $this->drupalPostForm(NULL, [
      'name' => 'admin',
      'mail' => 'admin@example.com',
      'subject[0][value]' => 'Message subject',
      'message[0][value]' => 'Message body',
    ], 'Send message');
    $this->dontSeeErrorMessage();
    $web_assert->pageTextContains('Form submitted');

    $this->drupalLogout();
  }

  /**
   * Test contact form with advanced settings.
   */
  private function _testContactFormWithAdditionalSettings() {
    $this->drupalLoginAsRoot();

    $this->drupalGet('/admin/structure/contact/add');
    $this->dontSeeErrorMessage();
    $web_assert = $this->assertSession();
    $web_assert->elementExists('css', '#edit-additional-settings');
    $web_assert->elementExists('css', '#edit-page-title');

    $contact_form_values = [
      'label' => 'Contact form with advanced settings',
      'id' => 'test_advanced_settings',
      'recipients' => 'test@example.com',
      'message' => 'Form submitted',
      'page_title' => 'New contact form page title',
      'name_label' => 'New name label',
      'email_label' => 'New e-mail label',
      'subject_label' => 'New subject label',
      'message_label' => 'New message label',
      'submit_button_text' => 'New submit button text',
      'email_subject' => 'New e-mail subject',
      'required_fields[name]' => FALSE,
      'required_fields[mail]' => FALSE,
      'required_fields[subject]' => FALSE,
      'required_fields[message]' => FALSE,
      'hide_preview' => TRUE,
    ];
    $this->drupalPostForm('/admin/structure/contact/add', $contact_form_values, 'Save');
    $this->dontSeeErrorMessage();
    $web_assert->pageTextContains('has been added');

    $this->drupalGet('/contact/' . $contact_form_values['id']);
    $this->dontSeeErrorMessage();
    $web_assert->elementTextContains('css', '#edit-name label', $contact_form_values['name_label']);
    $web_assert->elementTextContains('css', '#edit-mail label', $contact_form_values['email_label']);
    $web_assert->elementTextContains('css', '#edit-subject-wrapper label', $contact_form_values['subject_label']);
    $web_assert->elementTextContains('css', '#edit-message-wrapper label', $contact_form_values['message_label']);
    $web_assert->elementNotExists('css', '#edit-name.required');
    $web_assert->elementNotExists('css', '#edit-mail.required');
    $web_assert->elementNotExists('css', '#edit-subject-0-value.required');
    $web_assert->elementNotExists('css', '#edit-message-0-value.required');
    $web_assert->elementNotExists('css', '#edit-preview');
    $web_assert->elementAttributeContains('css', '#edit-submit', 'value', $contact_form_values['submit_button_text']);

    $this->drupalPostForm(NULL, [
      'name' => 'Dries',
      'mail' => 'dries@example.com',
      'subject[0][value]' => 'Message subject',
      'message[0][value]' => 'Message body',
    ], $contact_form_values['submit_button_text']);
    $this->dontSeeErrorMessage();
    $web_assert->pageTextContains('Form submitted');
    $this->assertMailString('subject', $contact_form_values['email_subject'], 1);

    $this->drupalLogout();
  }

  /**
   * Test render contact form.
   *
   * @covers \Drupal\improvements\ImprovementsHelper::getContactForm
   */
  private function _testContactFormRender() {
    $contact_form_entity = ContactForm::create([
      'id' => 'test_render_contact_form',
      'label' => 'Test render contact form',
      'recipients' => ['test@example.com'],
    ]);
    $contact_form_entity->save();

    $this->drupalLoginAsRoot();
    $this->drupalGet('/improvements-test/contact-form/' . $contact_form_entity->id());
    $web_assert = $this->assertSession();
    $web_assert->fieldExists('name');
    $web_assert->fieldExists('mail');
    $web_assert->fieldExists('message[0][value]');

    $this->deleteEntities($contact_form_entity);
    $this->drupalLogout();
  }

  /**
   * Test menu improvements.
   */
  private function _testMenuImprovements() {
    $this->drupalLoginAsRoot();
    $web_assert = $this->assertSession();

    // Test autofocus in menu form
    $this->drupalGet('/admin/structure/menu/add');
    $this->dontSeeErrorMessage();
    $web_assert->elementAttributeExists('css', '#edit-label', 'autofocus');

    // Test menu link form
    $this->drupalGet('/admin/structure/menu/manage/main/add');
    $this->dontSeeErrorMessage();
    $web_assert->elementAttributeExists('css', '#edit-title-0-value', 'autofocus');
    $web_assert->elementExists('css', '#edit-link-0-uri');
    $web_assert->elementNotExists('css', '#edit-link-0-uri[required]');

    $this->drupalLogout();
  }

  /**
   * Test "Add child item" in menu.
   */
  private function _testMenuLinkAddChildItem() {
    $this->drupalLoginAsRoot();
    $web_assert = $this->assertSession();

    $menu_link = MenuLinkContent::create([
      'title' => 'Home',
      'link' => ['uri' => 'internal:/'],
      'menu_name' => 'main',
    ]);
    $menu_link->save();
    $this->drupalGet('/admin/structure/menu/manage/main');
    $web_assert->elementExists('css', 'ul[data-drupal-selector="edit-links-menu-plugin-idmenu-link-content' . $menu_link->uuid() . '-operations"] li.add-child');
    $web_assert->elementAttributeContains('css', 'ul[data-drupal-selector="edit-links-menu-plugin-idmenu-link-content' . $menu_link->uuid() . '-operations"] li.add-child a', 'href', '/admin/structure/menu/manage/main/add?parent=main%3Amenu_link_content%3A' . $menu_link->uuid());

    $this->drupalGet('/admin/structure/menu/manage/main/add', ['query' => ['parent' => 'main:menu_link_content:' . $menu_link->uuid()]]);
    $web_assert->elementAttributeExists('css', 'option[value="main:menu_link_content:' . $menu_link->uuid() . '"]', 'selected');

    $menu_link->delete();
    $this->drupalLogout();
  }

  /**
   * Test menu link advanced settings.
   */
  private function _testMenuLinkAdvancedSettings() {
    $this->drupalLoginAsRoot();
    $web_assert = $this->assertSession();

    $this->drupalGet('/admin/structure/menu/manage/main/add');
    $web_assert->elementExists('css', '#edit-open-in-dialog');
    $web_assert->elementExists('css', '#edit-target-blank');
    $web_assert->elementExists('css', '#edit-item-class');
    $web_assert->elementExists('css', '#edit-link-class');

    $this->drupalPostForm(NULL, [
      'title[0][value]' => 'Test link',
      'link[0][uri]' => '/user',
      'open_in_dialog' => TRUE,
      'hide_href' => TRUE,
      'target_blank' => TRUE,
      'item_class' => 'test-link-item',
      'link_class' => 'test-link',
    ], 'Save');
    $this->dontSeeErrorMessage();
    $this->drupalGet('<front>');
    $web_assert->elementExists('css', '#block-main-menu-block li.test-link-item');
    $web_assert->elementExists('css', '#block-main-menu-block a.test-link');
    $web_assert->elementExists('css', '#block-main-menu-block a.test-link.use-ajax');
    $web_assert->elementAttributeContains('css', '#block-main-menu-block a.test-link', 'data-dialog-type', 'modal');
    $web_assert->elementAttributeContains('css', '#block-main-menu-block a.test-link', 'target', '_blank');
    $web_assert->elementAttributeContains('css', '#block-main-menu-block a.test-link', 'href', '#');
    $web_assert->elementAttributeContains('css', '#block-main-menu-block a.test-link', 'data-ajax-url', '/user');
    $web_assert->elementExists('css', 'script[src^="/core/misc/ajax.js"]');

    $this->drupalLogout();
  }

  /**
   * Test node improvements.
   */
  private function _testNodeImprovements() {
    $this->drupalLoginAsRoot();
    $web_assert = $this->assertSession();

    // Test "Node form" without advances settings
    $this->drupalGet('/node/add/page');
    $this->dontSeeErrorMessage();
    $web_assert->elementExists('css', '#edit-revision-information');
    $web_assert->elementExists('css', '.entity-content-form-footer .field--name-status');
    $web_assert->elementAttributeExists('css', '#edit-title-0-value', 'autofocus');

    // Test "Node type form"
    $this->drupalGet('/admin/structure/types/manage/page');
    $this->dontSeeErrorMessage();
    $web_assert->elementExists('css', '#edit-hide-revision-group');
    $web_assert->elementExists('css', '#edit-move-status-to-vertical-tab');
    $web_assert->checkboxChecked('menu_options[main]');

    // Test change advanced settings in "Node type form"
    $this->drupalPostForm(NULL, [
      'options[revision]' => FALSE,
      'hide_revision_group' => TRUE,
      'move_status_to_vertical_tab' => TRUE,
    ], 'Save content type');
    $this->dontSeeErrorMessage();
    $this->drupalGet('/admin/structure/types/manage/page');
    $this->dontSeeErrorMessage();
    $web_assert->checkboxChecked('hide_revision_group');
    $web_assert->checkboxChecked('move_status_to_vertical_tab');

    // Test "Node form" with advanced settings
    $this->drupalGet('/node/add/page');
    $this->dontSeeErrorMessage();
    $web_assert->elementNotExists('css', '#edit-revision-information');
    $web_assert->elementNotExists('css', '.entity-content-form-footer .field--name-status');
    $web_assert->elementExists('css', '#edit-options .field--name-status');

    // Test ""Node form" default settings
    $this->drupalGet('/admin/structure/types/add');
    $this->dontSeeErrorMessage();
    $web_assert->checkboxNotChecked('menu_options[main]');
    $web_assert->elementAttributeExists('css', '#edit-preview-mode-0', 'checked');
    $web_assert->checkboxNotChecked('options[promote]');
    $web_assert->checkboxNotChecked('options[revision]');
    $web_assert->checkboxNotChecked('display_submitted');

    $this->drupalLogout();
  }

  /**
   * Test condition plugin "term_depth".
   *
   * @covers \Drupal\improvements\Plugin\Condition\TermDepth
   */
  private function _testConditionTermDepth() {
    $condition_manager = $this->container->get('plugin.manager.condition'); /** @var ConditionManager $condition_manager */
    $term_depth_condition = $condition_manager->createInstance('term_depth'); /** @var ConditionInterface $term_depth_condition */
    $term1 = $this->createTerm($this->vocabulary, ['name' => 'Term 1 for test ' . __FUNCTION__]);
    $term1 = Term::load($term1->id());
    $term2 = $this->createTerm($this->vocabulary, ['name' => 'Term 2 for test ' . __FUNCTION__, 'parent' => [$term1->id()]]);
    $term2 = Term::load($term2->id());

    $term_depth_condition->setContextValue('taxonomy_term', $term1);
    $term_depth_condition->setConfig('depth', []);
    $this->assertIdentical(TRUE, $term_depth_condition->evaluate());

    $term_depth_condition->setContextValue('taxonomy_term', $term1);
    $term_depth_condition->setConfig('depth', [0]);
    $this->assertIdentical(TRUE, $term_depth_condition->evaluate());

    $term_depth_condition->setContextValue('taxonomy_term', $term1);
    $term_depth_condition->setConfig('depth', [0, 1, 2]);
    $this->assertIdentical(TRUE, $term_depth_condition->evaluate());

    $term_depth_condition->setContextValue('taxonomy_term', $term1);
    $term_depth_condition->setConfig('depth', [1]);
    $this->assertIdentical(FALSE, $term_depth_condition->evaluate());

    $term_depth_condition->setContextValue('taxonomy_term', $term2);
    $term_depth_condition->setConfig('depth', [0]);
    $this->assertIdentical(FALSE, $term_depth_condition->evaluate());

    $term_depth_condition->setContextValue('taxonomy_term', $term2);
    $term_depth_condition->setConfig('depth', [1]);
    $this->assertIdentical(TRUE, $term_depth_condition->evaluate());

    $term_depth_condition->setContextValue('taxonomy_term', $term2);
    $term_depth_condition->setConfig('depth', [2]);
    $this->assertIdentical(FALSE, $term_depth_condition->evaluate());

    $term_depth_condition->setContextValue('taxonomy_term', $term2);
    $term_depth_condition->setConfig('depth', [0, 2]);
    $this->assertIdentical(FALSE, $term_depth_condition->evaluate());

    $this->deleteEntities($term1, $term2);
  }

  /**
   * Test Pathauto "term depth" condition.
   */
  private function _testPathautoTermDepthCondition() {
    $this->drupalLoginAsRoot();
    $web_assert = $this->assertSession();

    $pathauto_pattern = PathautoPattern::create([
      'id' => 'term',
      'label' => 'Term',
      'type' => 'canonical_entities:taxonomy_term',
      'pattern' => 'term/[term:tid]',
    ]);
    $pathauto_pattern->save();

    $this->drupalGet('/admin/config/search/path/patterns/' . $pathauto_pattern->id());
    $this->dontSeeErrorMessage();
    $web_assert->pageTextContains('Term depth condition');
    $web_assert->elementExists('css', 'input[name="depth_condition[depth][0]"]');
    $web_assert->elementExists('css', 'input[name="depth_condition[depth][1]"]');
    $web_assert->elementExists('css', 'input[name="depth_condition[depth][2]"]');

    $this->drupalPostForm(NULL, [
      'depth_condition[depth][0]' => TRUE,
    ], 'Save');
    $this->dontSeeErrorMessage();
    $this->drupalGet('/admin/config/search/path/patterns/' . $pathauto_pattern->id());
    $this->dontSeeErrorMessage();
    $web_assert->checkboxChecked('depth_condition[depth][0]');

    $term1 = $this->createTerm($this->vocabulary);
    $term1 = Term::load($term1->id());
    $this->assertIdentical('/term/' . $term1->id(), $term1->get('path')->alias);

    $term2 = $this->createTerm($this->vocabulary, ['parent' => $term1->id()]);
    $term2 = Term::load($term2->id());
    $this->assertIdentical(NULL, $term2->get('path')->alias);

    $term1->delete();
    $term2->delete();
    $pathauto_pattern->delete();
    $this->drupalLogout();
  }

  /**
   * Test third party site information.
   */
  private function _testSiteInformationThirdPartySettings() {
    $this->drupalLoginAsRoot();
    $this->drupalGet('/admin/config/system/site-information');
    $this->dontSeeErrorMessage();
    $web_assert = $this->assertSession();
    $web_assert->fieldExists('mail_for_notifications');

    $this->drupalPostForm(NULL, [
      'mail_for_notifications' => 'notify@gmail.com',
    ], 'Save configuration');
    $this->drupalGet('/admin/config/system/site-information');
    $this->dontSeeErrorMessage();
    $web_assert->fieldValueEquals('mail_for_notifications', 'notify@gmail.com');

    $this->drupalLogout();
  }

  /**
   * Test "stored_depth" tem field.
   */
  private function _testTermDepth() {
    $term1 = $this->createTerm($this->vocabulary);
    $this->assertIdentical(0, $term1->get('stored_depth')->value);
    $this->assertIdentical(0, ImprovementsHelper::getTermDepth($term1));

    $term2 = $this->createTerm($this->vocabulary, [
      'parent' => [$term1->id()],
    ]);
    $this->assertIdentical(1, $term2->get('stored_depth')->value);
    $this->assertIdentical(1, ImprovementsHelper::getTermDepth($term2));

    $term3 = $this->createTerm($this->vocabulary, [
      'parent' => [$term2->id()],
    ]);
    $term4 = $this->createTerm($this->vocabulary, [
      'parent' => [$term2->id()],
    ]);
    $this->assertIdentical(2, $term3->get('stored_depth')->value);
    $this->assertIdentical(2, $term4->get('stored_depth')->value);
    $this->assertIdentical(2, ImprovementsHelper::getTermDepth($term3));
    $this->assertIdentical(2, ImprovementsHelper::getTermDepth($term3));

    $term2->set('parent', [0]);
    $term2->save();
    $this->assertIdentical(0, $term2->get('stored_depth')->value);
    $this->assertIdentical(0, ImprovementsHelper::getTermDepth($term2));
    $term3 = Term::load($term3->id()); // Load actual term
    $term4 = Term::load($term4->id()); // Load actual term
    $this->assertIdentical(1, (int)$term3->get('stored_depth')->value);
    $this->assertIdentical(1, (int)$term4->get('stored_depth')->value);
    $this->assertIdentical(1, ImprovementsHelper::getTermDepth($term3));
    $this->assertIdentical(1, ImprovementsHelper::getTermDepth($term4));

    $term1->delete();
    $term2->delete();
    $term3->delete();
    $term4->delete();
  }

  /**
   * Test helper function ImprovementsHelper::getTermChildrenCount().
   *
   * @covers \Drupal\improvements\ImprovementsHelper::getTermDirectChildrenCount
   */
  private function _testTermChildrenCount() {
    $term1 = $this->createTerm($this->vocabulary);
    $this->assertIdentical(0, ImprovementsHelper::getTermDirectChildrenCount($term1->id()));

    $term2 = $this->createTerm($this->vocabulary, ['parent' => [$term1->id()]]);
    $this->assertIdentical(0, ImprovementsHelper::getTermDirectChildrenCount($term2->id()));
    $this->assertIdentical(1, ImprovementsHelper::getTermDirectChildrenCount($term1->id()));

    $term3 = $this->createTerm($this->vocabulary, ['parent' => [$term1->id()]]);
    $this->assertIdentical(2, ImprovementsHelper::getTermDirectChildrenCount($term1->id()));

    $term3->delete();
    $this->assertIdentical(1, ImprovementsHelper::getTermDirectChildrenCount($term1->id()));

    $term2->delete();
    $this->assertIdentical(0, ImprovementsHelper::getTermDirectChildrenCount($term1->id()));

    $term1->delete();
  }

  /**
   * Test helper function ImprovementsHelper::getTermAllChildrenIds().
   *
   * @covers \Drupal\improvements\ImprovementsHelper::getTermAllChildrenIds
   */
  private function _testGetTermAllChildrenIds() {
    $term_1 = $this->createTerm($this->vocabulary, ['name' => 'Term 1']);
    $term_2 = $this->createTerm($this->vocabulary, ['name' => 'Term 2']);
    $this->assertIdentical([], ImprovementsHelper::getTermAllChildrenIds($term_1->bundle(), $term_1->id()));

    $term_1_1 = $this->createTerm($this->vocabulary, ['name' => 'Term 1.1', 'parent' => [$term_1->id()]]);
    $this->assertIdentical([], ImprovementsHelper::getTermAllChildrenIds($term_1_1->bundle(), $term_1_1->id()));
    $this->assertIdentical([
      (int)$term_1_1->id(),
    ], ImprovementsHelper::getTermAllChildrenIds($term_1->bundle(), $term_1->id()));

    $term_1_2 = $this->createTerm($this->vocabulary, ['name' => 'Term 1.2', 'parent' => [$term_1->id()]]);
    $this->assertIdentical([
      (int)$term_1_1->id(),
      (int)$term_1_2->id(),
    ], ImprovementsHelper::getTermAllChildrenIds($term_1->bundle(), $term_1->id()));

    $term_1_2_1 = $this->createTerm($this->vocabulary, ['name' => 'Term 1.2.1', 'parent' => [$term_1_2->id()]]);
    $this->assertIdentical([
      (int)$term_1_1->id(),
      (int)$term_1_2->id(),
      (int)$term_1_2_1->id(),
    ], ImprovementsHelper::getTermAllChildrenIds($term_1->bundle(), $term_1->id()));
    $this->assertIdentical([
      (int)$term_1_2_1->id(),
    ], ImprovementsHelper::getTermAllChildrenIds($term_1_2->bundle(), $term_1_2->id()));

    $term_1_2_1->delete();
    $this->assertIdentical([
      (int)$term_1_1->id(),
      (int)$term_1_2->id(),
    ], ImprovementsHelper::getTermAllChildrenIds($term_1->bundle(), $term_1->id()));

    $term_1_2->delete();
    $term_2->delete();
    $term_1_1->delete();
    $term_1->delete();
  }

  /**
   * Test "has_children" term field.
   */
  private function _testTermHasChildren() {
    $term1 = $this->createTerm($this->vocabulary);
    $this->assertIdentical(FALSE, $term1->get('has_children')->value);
    $this->assertIdentical(FALSE, ImprovementsHelper::termHasChildren($term1->id()));

    $term2 = $this->createTerm($this->vocabulary, ['parent' => [$term1->id()]]);
    $this->assertIdentical(FALSE, $term2->get('has_children')->value);
    $this->assertIdentical(FALSE, ImprovementsHelper::termHasChildren($term2->id()));
    $term1 = Term::load($term1->id()); // Load actual term
    $this->assertIdentical('1', $term1->get('has_children')->value);
    $this->assertIdentical(TRUE, ImprovementsHelper::termHasChildren($term1->id()));

    $term2->delete();
    $term1 = Term::load($term1->id()); // Load actual term
    $this->assertIdentical('0', $term1->get('has_children')->value);
    $this->assertIdentical(FALSE, ImprovementsHelper::termHasChildren($term1->id()));

    $term1->delete();
  }

  /**
   * Test taxonomy improvements.
   */
  private function _testTaxonomyImprovements() {
    $this->drupalLoginAsRoot();
    $web_assert = $this->assertSession();

    $this->drupalGet('/admin/structure/taxonomy/manage/' . $this->vocabulary->id() . '/add');
    $web_assert->elementAttributeExists('css', 'input[name="name[0][value]"]', 'autofocus');

    $term = $this->createTerm($this->vocabulary);
    $this->drupalGet('/admin/structure/taxonomy/manage/' . $this->vocabulary->id() . '/overview');
    $web_assert->elementNotExists('css', '#edit-reset-alphabetical');
    $web_assert->elementExists('css', '#edit-resave');

    $term->delete();
    $this->drupalLogout();
  }

  /**
   * Test link "Add child" term.
   *
   * @covers ::improvements_entity_operation
   */
  private function _testTermAddChildLink() {
    $this->drupalLoginAsRoot();
    $web_assert = $this->assertSession();

    $term = $this->createTerm($this->vocabulary);
    $this->drupalGet('/admin/structure/taxonomy/manage/' . $this->vocabulary->id() . '/overview');
    $web_assert->elementExists('css', 'tr[data-drupal-selector="edit-terms-tid' . $term->id() . '0"]');

    $this->drupalGet('/admin/structure/taxonomy/manage/' . $this->vocabulary->id() . '/add', ['query' => ['parent' => $term->id()]]);
    $web_assert->elementAttributeExists('css', 'select[name="parent[]"] option[value="' . $term->id() . '"]', 'selected');

    // Clean anvironment
    $term->delete();
    $this->drupalLogout();
  }

  /**
   * Test cookies_hash cache context.
   *
   * @covers \Drupal\improvements\Cache\CookiesHashCacheContext
   */
  private function _testCacheContextCookiesHash() {
    $request_stack = new RequestStack();
    $request = Request::create('/', 'GET', [], ['foo' => 'bar']);
    $request_stack->push($request);
    $cache_context = new CookiesHashCacheContext($request_stack);
    $this->assertSame('f4579c71', $cache_context->getContext());
    $this->assertSame('f4bb71a2', $cache_context->getContext('foo'));

    $request_stack = new RequestStack();
    $request = Request::create('/');
    $request_stack->push($request);
    $cache_context = new CookiesHashCacheContext($request_stack);
    $this->assertSame('', $cache_context->getContext());
    $this->assertSame('', $cache_context->getContext('foo'));
  }

  /**
   * Test entity_field cache context.
   *
   * @covers \Drupal\improvements\Cache\EntityFieldCacheContext
   */
  private function _testCacheContextEntityField() {
    $node = $this->createNode([
      'title' => 'Node for test ' . __FUNCTION__,
      'uid' => 1,
    ]);
    $route = new Route('/node/{node}');
    $route_match = new RouteMatch('entity.node.canonical', $route, ['node' => $node], ['node' => $node->id()]);
    $cache_context = new EntityFieldCacheContext($route_match);
    $this->assertIdentical((string)$node->get('uid')->target_id, $cache_context->getContext('uid'));
    $this->assertIdentical($node->label(), $cache_context->getContext('title'));
    $this->assertIdentical('', $cache_context->getContext('dummy'));

    $route = new Route('/');
    $route_match = new RouteMatch('<front>', $route);
    $cache_context = new EntityFieldCacheContext($route_match);
    $this->assertIdentical('', $cache_context->getContext('title'));
    $this->assertIdentical('', $cache_context->getContext('dummy'));

    $node->delete();
  }

  /**
   * Test entity_type cache context.
   *
   * @covers \Drupal\improvements\Cache\EntityTypeCacheContext
   */
  private function _testCacheContextEntityType() {
    $node = $this->createNode([
      'title' => 'Node for test ' . __FUNCTION__,
      'uid' => 1,
    ]);
    $route = new Route('/node/{node}');
    $route_match = new RouteMatch('entity.node.canonical', $route, ['node' => $node], ['node' => $node->id()]);
    $cache_context = new EntityTypeCacheContext($route_match);
    $this->assertIdentical('node', $cache_context->getContext());

    $route = new Route('/');
    $route_match = new RouteMatch('<front>', $route);
    $cache_context = new EntityTypeCacheContext($route_match);
    $this->assertIdentical('', $cache_context->getContext());

    $node->delete();
  }

  /**
   * Test is_entity_page cache context.
   *
   * @covers \Drupal\improvements\Cache\IsEntityPageCacheContext
   */
  private function _testCacheContextIsEntityPage() {
    $node = $this->createNode([
      'title' => 'Node for test ' . __FUNCTION__,
      'uid' => 1,
    ]);
    $route = new Route('/node/{node}');
    $route_match = new RouteMatch('entity.node.canonical', $route, ['node' => $node], ['node' => $node->id()]);
    $cache_context = new IsEntityPageCacheContext($route_match);
    $this->assertIdentical('1', $cache_context->getContext());
    $this->assertIdentical('1', $cache_context->getContext('node'));
    $this->assertIdentical('1', $cache_context->getContext('node,taxonomy_term'));
    $this->assertIdentical('0', $cache_context->getContext('taxonomy_term'));

    $route = new Route('/');
    $route_match = new RouteMatch('<front>', $route);
    $cache_context = new IsEntityPageCacheContext($route_match);
    $this->assertIdentical('0', $cache_context->getContext());
    $this->assertIdentical('0', $cache_context->getContext('node'));
    $this->assertIdentical('0', $cache_context->getContext('node,taxonomy_term'));
    $this->assertIdentical('0', $cache_context->getContext('taxonomy_term'));

    $node->delete();
  }

  /**
   * Test parent_term cache context.
   *
   * @covers \Drupal\improvements\Cache\ParentTermCacheContext
   */
  private function _testCacheContextParentTerm() {
    $term1 = $this->createTerm($this->vocabulary);
    $term2 = $this->createTerm($this->vocabulary, ['parent' => [$term1->id()]]);

    $route = new Route('/taxonomy/term/{taxonomy_term}');
    $route_match = new RouteMatch('entity.taxonomy_term.canonical', $route, ['taxonomy_term' => $term2], ['taxonomy_term' => $term2->id()]);
    $cache_context = new ParentTermCacheContext($route_match);
    $this->assertSame((string)$term1->id(), $cache_context->getContext());

    $term3 = $this->createTerm($this->vocabulary);
    $term2->set('parent', [$term1->id(), $term3->id()]);
    $term2->save();
    $this->assertSame($term1->id() . ',' . $term3->id(), $cache_context->getContext());

    $route = new Route('/');
    $route_match = new RouteMatch('<front>', $route);
    $cache_context = new ParentTermCacheContext($route_match);
    $this->assertSame('', $cache_context->getContext());
  }

  /**
   * Test Views field plugin "map".
   *
   * @covers \Drupal\improvements\Plugin\views\field\Map
   */
  private function _testViewsPluginFieldMap() {
    $node = $this->createNode([
      'title' => 'Node for test ' . __FUNCTION__,
      'test_map_field' => [
        'foo' => [
          'bar' => 'Hello world!',
        ],
      ],
    ]);
    $views_build = views_embed_view('test_views_map_field_plugin');
    $views_output = (string)$this->renderer->renderRoot($views_build);
    $this->assertStringContainsString('<span class="field-content">Hello world!</span>', $views_output);

    $this->deleteEntities($node);
  }

  /**
   * Test Views extended numeric filter.
   *
   * @covers \Drupal\improvements\Plugin\views\filter\ExtendedNumericFilter
   */
  private function _testViewsPluginFilterExtendedNumeric() {
    $this->drupalLoginAsRoot();
    $web_assert = $this->assertSession();

    $view = $this->createView([
      'id' => 'test_extended_numeric_filter',
      'label' => 'Test extended numeric filter',
      'base_table' => 'node_field_data',
    ]);
    $this->drupalPostForm('/admin/structure/views/nojs/add-handler/' . $view->id() . '/default/filter', ['name[node_field_data.nid]' => TRUE], 'Add and configure filter criteria');
    $this->dontSeeErrorMessage();
    $web_assert->elementExists('css', 'select[name="options[operator]"] option[value="in"]');

    $this->drupalPostForm(NULL, [
      'options[operator]' => 'in',
      'options[value][value]' => '1,2,3',
    ], 'Apply');
    $this->dontSeeErrorMessage();

    // @TODO Check generated sql query

    $this->drupalLogout();
  }

  /**
   * Test Views "is_current_user" sort plugin.
   *
   * @covers \Drupal\improvements\Plugin\views\sort\IsCurrentUser
   */
  private function _testViewsPluginSortIsCurrentUserSort() {
    $this->drupalLoginAsRoot();
    $sql = (string)$this->getViewQuery('test_sort_by_current_user');
    $this->assertStringContainsString('IF (node_field_data.uid = 1, ' . $this->rootUser->id() . ', 0) AS uid_is_current_user', $sql);
    $this->assertStringContainsString('ORDER BY uid_is_current_user', $sql);
  }

  /**
   * Test sort by comment author is current user
   */
  private function _testViewsPluginSortByCommentAuthorIsCurrentUser() {
    $this->drupalLoginAsRoot();
    $sql = (string)$this->getViewQuery('test_sort_by_comment_author_is_current_user');
    $this->assertStringContainsString('IF (comment_field_data.uid = 1, ' . $this->rootUser->id() . ', 0) AS uid_is_current_user', $sql);
    $this->assertStringContainsString('ORDER BY uid_is_current_user', $sql);
  }

  /**
   * Test Views "greater_than" sort plugin.
   *
   * @covers \Drupal\improvements\Plugin\views\sort\GreaterThan
   */
  private function _testViewsPluginSortGreaterThan() {
    $this->drupalLoginAsRoot();
    $sql = (string)$this->getViewQuery('test_sort_by_comment_author_is_registered');
    $this->assertStringContainsString('IF(comment_field_data.uid > 0, 1, 0) AS uid_greater_than', $sql);
    $this->assertStringContainsString('ORDER BY uid_greater_than', $sql);
  }

  /**
   * Test Views argument plugin "term_with_children".
   *
   * @covers \Drupal\improvements\Plugin\views\argument\TermWithChildrenArgument
   */
  private function _testViewsPluginArgumentTermWithChildren() {
    $this->drupalLoginAsRoot();

    $this->createField('node', 'page', 'field_category', 'entity_reference', 'Category', [
      'settings' => [
        'target_type' => 'taxonomy_term',
      ],
    ]);

    $this->drupalGet('/admin/structure/views/nojs/add-handler/test_argument_term_with_children/default/argument');
    $this->dontSeeErrorMessage();
    $web_assert = $this->assertSession();
    $web_assert->fieldExists('name[node__field_category.field_category_target_id_with_children]');

    $this->drupalPostForm(NULL, [
      'name[node__field_category.field_category_target_id_with_children]' => 'node__field_category.field_category_target_id_with_children',
    ], 'Add and configure contextual filters');
    $this->dontSeeErrorMessage();
    $this->drupalPostForm('/admin/structure/views/view/test_argument_term_with_children', [], 'Save');
    $this->dontSeeErrorMessage();

    $term_1 = $this->createTerm($this->vocabulary, ['name' => 'Term 1']);
    $term_2 = $this->createTerm($this->vocabulary, ['name' => 'Term 2']);
    $term_1_1 = $this->createTerm($this->vocabulary, ['name' => 'Term 1.1', 'parent' => [$term_1->id()]]);
    $term_1_2 = $this->createTerm($this->vocabulary, ['name' => 'Term 1.2', 'parent' => [$term_1->id()]]);

    $query = $this->getViewQuery('test_argument_term_with_children', 'default', [$term_1->id()]);
    $this->assertIdentical([
      (string)$term_1->id(),
      (string)$term_1_1->id(),
      (string)$term_1_2->id(),
    ], current($query->getArguments()));

    $query = $this->getViewQuery('test_argument_term_with_children', 'default', [$term_2->id()]);
    $this->assertIdentical((string)$term_2->id(), current($query->getArguments()));

    $this->deleteEntities($term_1, $term_2);
    $this->deleteField('field_category');
    $this->drupalLogout();
  }

  /**
   * Test token [view:total-rows-in-current-page].
   */
  private function _testViewsTotalRowsToken() {
    $node1 = $this->createNode(['title' => 'Node 1 for test ' . __FUNCTION__]);
    $node2 = $this->createNode(['title' => 'Node 2 for test ' . __FUNCTION__]);
    $node3 = $this->createNode(['title' => 'Node 3 for test ' . __FUNCTION__]);

    $view = $this->getExecutedView('test_total_rows_token');
    $total_rows = $this->tokenService->replace('[view:total-rows-in-current-page]', ['view' => $view]);
    $this->assertIdentical('3', $total_rows);

    $view = $this->getExecutedView('test_total_rows_token', 'default', [], function (ViewExecutable $view) {
      $view->setItemsPerPage(2);
    });
    $total_rows = $this->tokenService->replace('[view:total-rows-in-current-page]', ['view' => $view]);
    $this->assertIdentical('2', $total_rows);

    $view_build = views_embed_view('test_total_rows_token');
    $view_output = (string)$this->renderer->renderRoot($view_build);
    $this->assertStringContainsString('<div class="content-css-class total-rows-3">', $view_output);

    $this->deleteEntities($node1, $node2, $node3);
  }

  /**
   * Test hook hook_page_ROUTE_NAME_result_alter().
   *
   * @covers \Drupal\improvements\ImprovementsEventSubscriber::onKernelView
   */
  private function _testHookPageResultAlter() {
    $this->drupalGet('/improvements-test/altered-page-result');
    $web_assert = $this->assertSession();
    $web_assert->pageTextContains('Altered result');
    $web_assert->pageTextNotContains('Markup from controller');
  }

  /**
   * Test twig function entity_url().
   *
   * @covers \Drupal\improvements\ImprovementsTwigExtension::entityUrl
   */
  private function _testTwigFunctionEntityUrl() {
    $node = $this->createNode(['title' => 'Node for test ' . __FUNCTION__]);
    $build = [
      '#type' => 'inline_template',
      '#template' => '{{ entity_url(node) }}',
      '#context' => [
        'node' => $node,
      ],
    ];
    $output = (string)$this->renderer->renderRoot($build);
    $this->assertIdentical('/node/' . $node->id(), $output);

    $this->deleteEntities($node);
  }

  /**
   * Test twig function preg_matches().
   *
   * @covers \Drupal\improvements\ImprovementsTwigExtension::pregMatches
   */
  private function _testTwigFunctionPregMatch() {
    $build = [
      '#type' => 'inline_template',
      '#template' => "{{ preg_matches('/[0-9]+/', 'Drupal 9', 0) }}",
    ];
    $output = (string)$this->renderer->renderRoot($build);
    $this->assertIdentical('9', $output);

    $build = [
      '#type' => 'inline_template',
      '#template' => "{% if preg_matches('/Drupal [0-9]+/', 'Drupal 9') %}ok{% endif %}",
    ];
    $output = (string)$this->renderer->renderRoot($build);
    $this->assertIdentical('ok', $output);
  }

  /**
   * Test twig filter format_size.
   */
  private function _testTwigFilterFormatSize() {
    $build = [
      '#type' => 'inline_template',
      '#template' => "{{ 123|format_size }}",
    ];
    $output = (string)$this->renderer->renderRoot($build);
    $this->assertIdentical('123 bytes', $output);
  }

  /**
   * Test filter plugin "nofollow_external_links_filter".
   *
   * @covers \Drupal\improvements\Plugin\Filter\NofollowExternalLinksFilter
   */
  private function _testFilterPluginNofollowExternalLinks() {
    $filter = $this->createFilterFormat([
      'format' => 'test_nofollow_external_links',
      'name' => 'Test nofollow_external_links_filter',
      'filters' => [
        'nofollow_external_links_filter' => [
          'status' => 1,
        ],
      ],
    ]);
    $this->drupalLoginAsRoot();
    $this->drupalGet('/admin/config/content/formats/manage/' . $filter->id());
    $this->dontSeeErrorMessage();
    $web_assert = $this->assertSession();
    $web_assert->fieldExists('filters[nofollow_external_links_filter][status]');
    $web_assert->checkboxChecked('filters[nofollow_external_links_filter][status]');

    $result = check_markup('foo <a href="https://google.com">Google</a> bar', $filter->id());
    $this->assertIdentical('foo <a href="https://google.com" rel="nofollow">Google</a> bar', (string)$result);

    $this->deleteEntities($filter);
    $this->drupalLogout();
  }

  /**
   * Test filter plugin "table_wrapper_filter".
   *
   * @covers \Drupal\improvements\Plugin\Filter\TableWrapperFilter
   */
  private function _testFilterPluginTableWrapper() {
    $filter = $this->createFilterFormat([
      'format' => 'table_wrapper_filter',
      'name' => 'Test table_wrapper_filter',
      'filters' => [
        'table_wrapper_filter' => [
          'status' => 1,
        ],
      ],
    ]);
    $this->drupalLoginAsRoot();
    $this->drupalGet('/admin/config/content/formats/manage/' . $filter->id());
    $this->dontSeeErrorMessage();
    $web_assert = $this->assertSession();
    $web_assert->fieldExists('filters[table_wrapper_filter][status]');
    $web_assert->checkboxChecked('filters[table_wrapper_filter][status]');

    $result = check_markup('foo <table></table> bar', $filter->id());
    $this->assertIdentical('foo <div class="table-wrapper"><table></table></div> bar', (string)$result);

    $this->deleteEntities($filter);
    $this->drupalLogout();
  }

  /**
   * Test access check by uuid.
   *
   * @covers \Drupal\improvements\Access\UuidAccessCheck
   */
  private function _testAccessCheckUuid() {
    $node = $this->createNode(['title' => 'Node for test ' . __FUNCTION__]);

    $this->drupalGet('/improvements-test/uuid-access-check/' . $node->id());
    $this->assertResponse(403);

    $this->drupalGet('/improvements-test/uuid-access-check/' . $node->id(), ['query' => ['uuid' => '123']]);
    $this->assertResponse(403);

    $this->drupalGet('/improvements-test/uuid-access-check/' . $node->id(), ['query' => ['uuid' => 'true']]);
    $this->assertResponse(403);

    $this->drupalGet('/improvements-test/uuid-access-check/' . $node->id(), ['query' => ['uuid' => $node->uuid()]]);
    $this->assertResponse(200);

    $this->deleteEntities($node);
  }

  /**
   * Test token [current-page:url-with-query]
   */
  private function _testTokenCurrentPageUrlWithQuery() {
    $request = Request::create('/admin', 'GET', ['foo' => 'bar']);
    $request_stack = $this->container->get('request_stack');
    $request_stack->push($request);
    $this->assertStringContainsString('/admin?foo=bar', $this->tokenService->replace('[current-page:url-with-query]'));
    $this->assertIdentical('/admin?foo=bar', $this->tokenService->replace('[current-page:url-with-query:relative]'));
    $request_stack->pop();
  }

  /**
   * Test tokens [site:third_party_setting_name].
   */
  private function _testTokenSiteThirdPartySettings() {
    $this->configFactory->getEditable('improvements.site_third_party_settings')
      ->set('settings.mail_for_notifications', 'mail-for-notifications@example.com')
      ->save();
    $this->assertIdentical('mail-for-notifications@example.com', $this->tokenService->replace('[site:mail_for_notifications]'));
  }

  /**
   * Open field widget/formatter settings.
   */
  private function openEntityDisplayComponentSettings(string $field_name) {
    $this->drupalPostForm(NULL, [], $field_name . '_settings_edit');
    $this->dontSeeErrorMessage();
  }

}
